pointer_header = "<memory>"          # which C++ header should be included
pointer_type = "std::shared_ptr<T>"  # the name of the smart pointer parameterized on T
get_pointer = "(ptr.get())"          # an expression that converts a smart pointer named ptr to T *
max_args = 4                         # the maximum number of parameters to be processed

print """#ifndef AS_SMART_POINTER_WRAPPER
#define AS_SMART_POINTER_WRAPPER

#include <new>
#include "angelscript.h"
#include {0}
""".format(pointer_header)

print """namespace spw {

template <typename T> struct Default { 
  static T f(void) { return T(); } 
};
template <typename T> struct Default<T &> { 
  static T & f(void) { return *static_cast<T *>(0); } 
};

template <typename T> class Proxy {
  public:
    T value;
    Proxy(T value) : value(value) {}
  private:
    Proxy(const Proxy &);
    Proxy & operator=(const Proxy &);
};
"""

print """template <typename T, typename U, U T::*member_ptr>
U getter(const {0} & ptr) {{
  if (!ptr) {{
    AS_NAMESPACE_QUALIFIER asIScriptContext * ctx = AS_NAMESPACE_QUALIFIER asGetActiveContext();
    ctx->SetException("Attempting to access member of a null pointer");
    return Default<U>::f();
  }}
  return {1}->*member_ptr;
}}

template <typename T, typename U, U T::*member_ptr>
const U & ref_getter(const {0} & ptr) {{
  if (!ptr) {{
    AS_NAMESPACE_QUALIFIER asIScriptContext * ctx = AS_NAMESPACE_QUALIFIER asGetActiveContext();
    ctx->SetException("Attempting to access member of a null pointer");
    return Default<U &>::f();
  }}
  return {1}->*member_ptr;
}}

template <typename T, typename U, U T::*member_ptr>
void getter_gen(AS_NAMESPACE_QUALIFIER asIScriptGeneric * gen) {{
  {0} & ptr = *static_cast<{0} *>(gen->GetObject());
  if (!ptr) {{
    AS_NAMESPACE_QUALIFIER asIScriptContext * ctx = AS_NAMESPACE_QUALIFIER asGetActiveContext();
    ctx->SetException("Attempting to access member of a null pointer");
  }} else {{
    new (gen->GetAddressOfReturnLocation()) Proxy<U>({1}->*member_ptr);
  }}
}}

template <typename T, typename U, U T::*member_ptr>
void ref_getter_gen(AS_NAMESPACE_QUALIFIER asIScriptGeneric * gen) {{
  {0} & ptr = *static_cast<{0} *>(gen->GetObject());
  if (!ptr) {{
    AS_NAMESPACE_QUALIFIER asIScriptContext * ctx = AS_NAMESPACE_QUALIFIER asGetActiveContext();
    ctx->SetException("Attempting to access member of a null pointer");
  }} else {{
    new (gen->GetAddressOfReturnLocation()) Proxy<U &>({1}->*member_ptr);
  }}
}}

template <typename T, typename U, U T::*member_ptr>
void setter(const {0} & ptr, U value) {{
  if (!ptr) {{
    AS_NAMESPACE_QUALIFIER asIScriptContext * ctx = AS_NAMESPACE_QUALIFIER asGetActiveContext();
    ctx->SetException("Attempting to access member of a null pointer");
  }} else {{
    {1}->*member_ptr = value;
  }}
}}

template <typename T, typename U, U T::*member_ptr>
void ref_setter(const {0} & ptr, const U & value) {{
  if (!ptr) {{
    AS_NAMESPACE_QUALIFIER asIScriptContext * ctx = AS_NAMESPACE_QUALIFIER asGetActiveContext();
    ctx->SetException("Attempting to access member of a null pointer");
  }} else {{
    {1}->*member_ptr = value;
  }}
}}

template <typename T, typename U, U T::*member_ptr>
void setter_gen(AS_NAMESPACE_QUALIFIER asIScriptGeneric * gen) {{
  {0} & ptr = *static_cast<{0} *>(gen->GetObject());
  if (!ptr) {{
    AS_NAMESPACE_QUALIFIER asIScriptContext * ctx = AS_NAMESPACE_QUALIFIER asGetActiveContext();
    ctx->SetException("Attempting to access member of a null pointer");
  }} else {{
    {1}->*member_ptr = static_cast<Proxy<U> *>(gen->GetAddressOfArg(0))->value;
  }}
}}

template <typename T, typename U, U T::*member_ptr>
void ref_setter_gen(AS_NAMESPACE_QUALIFIER asIScriptGeneric * gen) {{
  {0} & ptr = *static_cast<{0} *>(gen->GetObject());
  if (!ptr) {{
    AS_NAMESPACE_QUALIFIER asIScriptContext * ctx = AS_NAMESPACE_QUALIFIER asGetActiveContext();
    ctx->SetException("Attempting to access member of a null pointer");
  }} else {{
    {1}->*member_ptr = static_cast<Proxy<U &> *>(gen->GetAddressOfArg(0))->value;
  }}
}}
""".format(pointer_type, get_pointer)

print """template <typename T> struct Caller {};
template <typename T> struct GenCaller {};
"""

typename_list = ", typename A0"
type_list = "A0"
parameter_list = ", A0 a0"
arg_list = "a0"
gen_arg_list = "\n        static_cast<Proxy <A0> *>(gen->GetAddressOfArg(0))->value"

caller = """template <typename T, typename R{0}>
struct Caller<R (T::*)({1}){6}> {{
  template <R (T::*fn)({1}){6}>
  static R f({2} & ptr{3}) {{
    if (!ptr) {{
      AS_NAMESPACE_QUALIFIER asIScriptContext * ctx = AS_NAMESPACE_QUALIFIER asGetActiveContext();
      ctx->SetException("Attempting to call member function of a null pointer");
      return Default<R>::f();
    }} else {{
      return ({4}->*fn)({5});
    }}
  }}
}};
"""

gen_caller = """template <typename T{0}>
struct GenCaller<void (T::*)({1}){6}> {{
  template <void (T::*fn)({1}){6}>
  static void f(AS_NAMESPACE_QUALIFIER asIScriptGeneric * gen) {{
    {2} & ptr = *static_cast<{2} *>(gen->GetObject());
    if (!ptr) {{
      AS_NAMESPACE_QUALIFIER asIScriptContext * ctx = AS_NAMESPACE_QUALIFIER asGetActiveContext();
      ctx->SetException("Attempting to call member function of a null pointer");
    }} else {{
      ({4}->*fn)({5});
    }}
  }}
}};

template <typename T, typename R{0}>
struct GenCaller<R (T::*)({1}){6}> {{
  template <R (T::*fn)({1}){6}>
  static void f(AS_NAMESPACE_QUALIFIER asIScriptGeneric * gen) {{
    {2} & ptr = *static_cast<{2} *>(gen->GetObject());
    if (!ptr) {{
      AS_NAMESPACE_QUALIFIER asIScriptContext * ctx = AS_NAMESPACE_QUALIFIER asGetActiveContext();
      ctx->SetException("Attempting to call member function of a null pointer");
    }} else {{
      new (gen->GetAddressOfReturnLocation()) Proxy<R>(({4}->*fn)({5}));
    }}
  }}
}};
"""

print caller.format("", "void", pointer_type, "", get_pointer, "", "")
print caller.format("", "void", pointer_type, "", get_pointer, "", " const")
print gen_caller.format("", "void", pointer_type, "", get_pointer, "", "")
print gen_caller.format("", "void", pointer_type, "", get_pointer, "", " const")

print caller.format(typename_list, type_list, pointer_type, parameter_list, get_pointer, arg_list, "")
print caller.format(typename_list, type_list, pointer_type, parameter_list, get_pointer, arg_list, " const")
print gen_caller.format(typename_list, type_list, pointer_type, parameter_list, get_pointer, gen_arg_list, "")
print gen_caller.format(typename_list, type_list, pointer_type, parameter_list, get_pointer, gen_arg_list, " const")

for i in range(1, max_args):
    typename_list += ", typename A{0}".format(i)
    type_list += ", A{0}".format(i)
    parameter_list += ", A{0} a{0}".format(i)
    arg_list += ", a{0}".format(i)
    gen_arg_list += ",\n        static_cast<Proxy <A{0}> *>(gen->GetAddressOfArg({0}))->value".format(i)

    print caller.format(typename_list, type_list, pointer_type, parameter_list, get_pointer, arg_list, "")
    print caller.format(typename_list, type_list, pointer_type, parameter_list, get_pointer, arg_list, " const")
    print gen_caller.format(typename_list, type_list, pointer_type, parameter_list, get_pointer, gen_arg_list, "")
    print gen_caller.format(typename_list, type_list, pointer_type, parameter_list, get_pointer, gen_arg_list, " const")

print """template <typename T, typename U>
struct GetSet {
  template <U T::* member_ptr> AS_NAMESPACE_QUALIFIER asSFuncPtr   g(void) { return asFUNCTION((    &getter    <T, U, member_ptr>)); }
  template <U T::* member_ptr> AS_NAMESPACE_QUALIFIER asSFuncPtr  rg(void) { return asFUNCTION((&ref_getter    <T, U, member_ptr>)); }
  template <U T::* member_ptr> AS_NAMESPACE_QUALIFIER asSFuncPtr  gg(void) { return asFUNCTION((    &getter_gen<T, U, member_ptr>)); }
  template <U T::* member_ptr> AS_NAMESPACE_QUALIFIER asSFuncPtr rgg(void) { return asFUNCTION((&ref_getter_gen<T, U, member_ptr>)); }

  template <U T::* member_ptr> AS_NAMESPACE_QUALIFIER asSFuncPtr   s(void) { return asFUNCTION((&setter        <T, U, member_ptr>)); }
  template <U T::* member_ptr> AS_NAMESPACE_QUALIFIER asSFuncPtr  rs(void) { return asFUNCTION((&ref_setter    <T, U, member_ptr>)); }
  template <U T::* member_ptr> AS_NAMESPACE_QUALIFIER asSFuncPtr  sg(void) { return asFUNCTION((    &setter_gen<T, U, member_ptr>)); }
  template <U T::* member_ptr> AS_NAMESPACE_QUALIFIER asSFuncPtr rsg(void) { return asFUNCTION((&ref_setter_gen<T, U, member_ptr>)); }
};

template <typename T, typename U>
GetSet<T, U> gs_id(U T::*) { return GetSet<T, U>(); }

#define GETTER(ClassName, member_name)         (::spw::gs_id(&ClassName::member_name).g<&ClassName::member_name>())
#define REF_GETTER(ClassName, member_name)     (::spw::gs_id(&ClassName::member_name).rg<&ClassName::member_name>())
#define GETTER_GEN(ClassName, member_name)     (::spw::gs_id(&ClassName::member_name).gg<&ClassName::member_name>())
#define REF_GETTER_GEN(ClassName, member_name) (::spw::gs_id(&ClassName::member_name).rgg<&ClassName::member_name>())

#define     SETTER(ClassName, member_name)     (::spw::gs_id(&ClassName::member_name).s<&ClassName::member_name>())
#define REF_SETTER(ClassName, member_name)     (::spw::gs_id(&ClassName::member_name).rs<&ClassName::member_name>())
#define     SETTER_GEN(ClassName, member_name) (::spw::gs_id(&ClassName::member_name).sg<&ClassName::member_name>())
#define REF_SETTER_GEN(ClassName, member_name) (::spw::gs_id(&ClassName::member_name).rsg<&ClassName::member_name>())

template <typename T>
struct CCaller {
  template <T ptr> AS_NAMESPACE_QUALIFIER asSFuncPtr c(void) { return asFUNCTION(&Caller<T>::template f<ptr>); }
  template <T ptr> AS_NAMESPACE_QUALIFIER asSFuncPtr g(void) { return asFUNCTION(&GenCaller<T>::template f<ptr>); }
};

template <typename T>
CCaller<T> caller_id(T) { return CCaller<T>(); }

#define     CALLER(ClassName, member_name)                  (::spw::caller_id(&ClassName::member_name).c< &ClassName::member_name >())
#define GEN_CALLER(ClassName, member_name)                  (::spw::caller_id(&ClassName::member_name).g< &ClassName::member_name >())
#define     CALLER_PR(ClassName, member_name, Args, Return) asFUNCTION((::spw::Caller<Return (ClassName::*)Args>::f< &ClassName::member_name >))
#define GEN_CALLER_PR(ClassName, member_name, Args, Return) asFUNCTION((::spw::GenCaller<Return (ClassName::*)Args>::f< &ClassName::member_name >))

} // end namespace spw

#endif"""